from .report import Record, FileNotFound, ReportInvalidOrder

def read_abbr(path, records_dict=None, abbr_file="abbreviations.txt"):
    return Record.read_abbr(path, records_dict, abbr_file)

def read_logs(path, records_dict=None, start_file="start.log", end_file="end.log"):
    return Record.read_logs(path, records_dict, start_file, end_file)

def build_report(path, order="asc"):
    return Record.build_report(path, order)

def print_report(good_records, bad_records, border_line: int = 15):
    return Record.print_report(good_records, bad_records)

def cli(args_list=None):
    return Record.cli(args_list)

__all__ = ["FileNotFound", "ReportInvalidOrder", "Record", "cli", "build_report", "read_abbr", "read_logs", "print_report"]


from argparse import Namespace
from src.report_vp.report import Record
import pytest
from datetime import datetime
import pathlib


@pytest.fixture
def mock_data_dir(tmp_path: pathlib.Path):
    """
    Creates a temporary directory with mock files for testing
    """
    mock_dir = tmp_path  # No need to create a subdirectory

    (mock_dir / "abbreviations.txt").write_text(
        "CLS_Charles Leclerc_SAUBER FERRARI\nLHM_Lewis Hamilton_MERCEDES\nDRR_Daniel Ricciardo_RED BULL RACING TAG HEUER\nABC\n1233@@abc"
    )
    (mock_dir / "start.log").write_text(
        "CLS2018-05-24_11:10:54.750\nLHM2018-05-24_12:10:32.585\nAHM2018-05-24_12:10:32.585\n_____\nDRR2018-05-24_12:14:12.054\n"
    )
    (mock_dir / "end.log").write_text(
        "CLS2018-05-24_12:10:54.750\nLHM2018-05-24_12:11:32.585\nAHM2018-05-24_12:11:32.585\n_____\nDRR2018-05-24_12:11:24.067"
    )

    return mock_dir


@pytest.mark.usefixtures("mock_data_dir")
def test_read_abbr(mock_data_dir):
    folder_path = mock_data_dir
    records_dict = {}
    abbr_file = "abbreviations.txt"

    output = Record.read_abbr(folder_path, records_dict, abbr_file)

    expected_data = [
        ("CLS", "Charles Leclerc", "SAUBER FERRARI"),
        ("LHM", "Lewis Hamilton", "MERCEDES"),
        ("DRR", "Daniel Ricciardo", "RED BULL RACING TAG HEUER"),
    ]

    for abbr, driver, team in expected_data:
        assert abbr in output
        assert output[abbr].driver == driver
        assert output[abbr].team == team

    assert "Bad data: abbreviations.txt, line: 4" in output
    assert "Bad data: abbreviations.txt, line: 5" in output


@pytest.mark.usefixtures("mock_data_dir")
def test_read_logs(mock_data_dir):
    folder_path = mock_data_dir

    mock_records_dict = {
        "CLS": Record(),
        "LHM": Record(),
        "DRR": Record(),
    }

    output_dict = Record.read_logs(
        folder_path, mock_records_dict, start_file="start.log", end_file="end.log"
    )

    expected_valid_cases = {
        "CLS": Record(
            start=datetime(2018, 5, 24, 11, 10, 54, 750000),
            end=datetime(2018, 5, 24, 12, 10, 54, 750000),
            error=None,
        ),
        "LHM": Record(
            start=datetime(2018, 5, 24, 12, 10, 32, 585000),
            end=datetime(2018, 5, 24, 12, 11, 32, 585000),
            error=None,
        ),
        "DRR": Record(
            start=datetime(2018, 5, 24, 12, 14, 12, 54000),
            end=None,
            error="Invalid end time: must be after the start time",
        ),
    }

    expected_invalid_cases = {
        "Bad abbr: AHM, line: 3, file: start.log": Record(
            error="Bad abbr: AHM, line: 3, file: start.log"
        ),
        "Bad data: start.log, line: 4": Record(error="Bad data: start.log, line: 4"),
        "Bad abbr: AHM, line: 3, file: end.log": Record(
            error="Bad abbr: AHM, line: 3, file: end.log"
        ),
        "Bad data: end.log, line: 4": Record(error="Bad data: end.log, line: 4"),
    }

    for record_key, expected_record in expected_valid_cases.items():
        assert output_dict[record_key] == expected_record

    for error_key, expected_error_record in expected_invalid_cases.items():
        assert output_dict[error_key].error == expected_error_record.error


@pytest.mark.usefixtures("mock_data_dir")
def test_build_report(mock_data_dir):
    folder_path = mock_data_dir

    # Expected Cases
    expected_good_records_asc = [
        Record(
            driver="Lewis Hamilton",
            team="MERCEDES",
            start=datetime(2018, 5, 24, 12, 10, 32, 585000),
            end=datetime(2018, 5, 24, 12, 11, 32, 585000),
            error=None,
        ),
        Record(
            driver="Charles Leclerc",
            team="SAUBER FERRARI",
            start=datetime(2018, 5, 24, 11, 10, 54, 750000),
            end=datetime(2018, 5, 24, 12, 10, 54, 750000),
            error=None,
        ),
    ]
    expected_good_records_desc = [
        Record(
            driver="Charles Leclerc",
            team="SAUBER FERRARI",
            start=datetime(2018, 5, 24, 11, 10, 54, 750000),
            end=datetime(2018, 5, 24, 12, 10, 54, 750000),
            error=None,
        ),
        Record(
            driver="Lewis Hamilton",
            team="MERCEDES",
            start=datetime(2018, 5, 24, 12, 10, 32, 585000),
            end=datetime(2018, 5, 24, 12, 11, 32, 585000),
            error=None,
        ),
    ]

    expected_bad_records = [
        Record(
            driver="Daniel Ricciardo",
            team="RED BULL RACING TAG HEUER",
            start=datetime(2018, 5, 24, 12, 14, 12, 54000),
            end=None,
            error="Invalid end time: must be after the start time",
        ),
        Record(
            driver=None,
            team=None,
            start=None,
            end=None,
            error="Bad data: abbreviations.txt, line: 4",
        ),
        Record(
            driver=None,
            team=None,
            start=None,
            end=None,
            error="Bad data: abbreviations.txt, line: 5",
        ),
        Record(
            driver=None,
            team=None,
            start=None,
            end=None,
            error="Bad abbr: AHM, line: 3, file: start.log",
        ),
        Record(
            driver=None,
            team=None,
            start=None,
            end=None,
            error="Bad data: start.log, line: 4",
        ),
        Record(
            driver=None,
            team=None,
            start=None,
            end=None,
            error="Bad abbr: AHM, line: 3, file: end.log",
        ),
        Record(
            driver=None,
            team=None,
            start=None,
            end=None,
            error="Bad data: end.log, line: 4",
        ),
    ]

    # Test Case 1: Ascending Order
    good_records_asc, bad_records = Record.build_report(path=folder_path, order="asc")

    assert good_records_asc == expected_good_records_asc
    assert bad_records == expected_bad_records

    # Test Case 2: Descending Order
    good_records_desc, bad_records = Record.build_report(path=folder_path, order="desc")

    assert good_records_desc == expected_good_records_desc
    assert bad_records == expected_bad_records

    # Test Case 3: Invalid Order
    good_records_error, bad_records_error = Record.build_report(
        path=folder_path, order="error"
    )

    assert good_records_error == []
    assert bad_records_error == []


def test_print_report():
    expected_good_records_asc = [
        Record(
            driver="Lewis Hamilton",
            team="MERCEDES",
            start=datetime(2018, 5, 24, 12, 10, 32, 585000),
            end=datetime(2018, 5, 24, 12, 11, 32, 585000),
            error=None,
        ),
        Record(
            driver="Charles Leclerc",
            team="SAUBER FERRARI",
            start=datetime(2018, 5, 24, 11, 10, 54, 750000),
            end=datetime(2018, 5, 24, 12, 10, 54, 750000),
            error=None,
        ),
    ]

    expected_bad_records = [
        Record(
            driver="Daniel Ricciardo",
            team="RED BULL RACING TAG HEUER",
            start=datetime(2018, 5, 24, 12, 14, 12, 54000),
            end=None,
            error="Invalid end time: must be after the start time",
        ),
        Record(
            driver=None,
            team=None,
            start=None,
            end=None,
            error="Bad data: abbreviations.txt, line: 4",
        ),
        Record(
            driver=None,
            team=None,
            start=None,
            end=None,
            error="Bad data: abbreviations.txt, line: 5",
        ),
        Record(
            driver=None,
            team=None,
            start=None,
            end=None,
            error="Bad abbr: AHM, line: 3, file: start.log",
        ),
        Record(
            driver=None,
            team=None,
            start=None,
            end=None,
            error="Bad data: start.log, line: 4",
        ),
    ]

    print_report = Record.print_report(
        expected_good_records_asc, expected_bad_records, border_line=2
    )

    assert (
        print_report
        == """ 1. Lewis Hamilton       | MERCEDES                       | 0:01:00
 2. Charles Leclerc      | SAUBER FERRARI                 | 1:00:00
---------------------------------------------------------------------------
 3. Daniel Ricciardo     | RED BULL RACING TAG HEUER      | Invalid end time: must be after the start time
 4. Missing Data         | Missing Data                   | Bad data: abbreviations.txt, line: 4
 5. Missing Data         | Missing Data                   | Bad data: abbreviations.txt, line: 5
 6. Missing Data         | Missing Data                   | Bad abbr: AHM, line: 3, file: start.log
 7. Missing Data         | Missing Data                   | Bad data: start.log, line: 4
"""
    )


@pytest.mark.usefixtures("mock_data_dir")
def test_cli(mock_data_dir):
    dir_path = str(mock_data_dir)
    args_list = [
        ["--files", dir_path, "--asc"],
        ["--files", dir_path, "--desc"],
        ["--files", dir_path, "--asc", "--driver", "John Doe"],
    ]

    expected_outputs = [
        Namespace(files=dir_path, asc=True, desc=False, driver=None, sort="asc"),
        Namespace(files=dir_path, asc=False, desc=True, driver=None, sort="desc"),
        Namespace(files=dir_path, asc=True, desc=False, driver="John Doe", sort="asc"),
    ]

    outputs = []

    for args in args_list:
        output = Record.cli(args)
        outputs.append(output)
        print

    for output, expected_output in zip(outputs, expected_outputs):
        assert output == expected_output


if __name__ == "__main__":
    pytest.main([f"{__file__}", "-rx", "-vv"])



import argparse
from datetime import datetime, timedelta
from pathlib import Path
import re
import copy


class FileNotFound(Exception):
    pass


class ReportInvalidOrder(Exception):
    pass


class Record:
    regex_abbr = re.compile(r"(^[A-Z]{3})_(\w+\s\w+)_([A-Z\s]+$)")
    regex_log = re.compile(r"(^[A-Z]{3})(\d+-\d+-\d+_\d+:\d+:\d+.\d+)")

    def __init__(
        self,
        abbr: str = None,
        driver: str = None,
        team: str = None,
        start: datetime = None,
        end: datetime = None,
        error: str = None,
    ):
        self.abbr = abbr
        self.driver = driver
        self.team = team
        self._start = start
        self._end = end
        self.error = error

    @property
    def start(self):
        return self._start

    @start.setter
    def start(self, value):
        if isinstance(value, datetime):
            if not self.end or self.end > value:
                self._start = value
            else:
                self.error = "Invalid start time: must be before the end time"
        else:
            self.error = "start() expects a datetime object"

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, value):
        if isinstance(value, datetime):
            if not self.start or self.start < value:
                self._end = value
            else:
                self.error = "Invalid end time: must be after the start time"
        else:
            self.error = "end() expects a datetime object"

    @property
    def result(self):
        return self.end - self.start if self.start and self.end else None

    def __eq__(self, value: object) -> bool:
        if isinstance(value, Record):
            return (
                self.abbr == value.abbr
                and self.team == value.team
                and self.driver == value.driver
                and self._end == value.end
                and self._start == value.start
                and self.error == value.error
            )
        else:
            return NotImplemented

    def __ne__(self, value: object) -> bool:
        x = self.__eq__(value)
        return not x if x is not NotImplemented else NotImplemented

    def __repr__(self):
        return f"Record({self.driver}, {self.team}, {self.start}, {self.end}, {self.error})"

    def __str__(self):
        return f"{self.driver:<20} | {self.team:<30} | {self.result if not self.error else self}"

    @staticmethod
    def read_abbr(
        path: str | Path,
        records_dict: dict[str, "Record"] = None,
        abbr_file: str = "abbreviations.txt",
    ) -> dict[str, "Record"]:
        """
        Validate the file path, extract abbreviations, and save them in a deep copy of the data
        dictionary, retaining only lines that match the regular expression pattern.
        Args:
            path: Path to the directory containing abbreviations file.
            records_dict: Optional dictionary to populate (defaults to empty).
            abbr_file: Name of the abbreviations file (defaults to "abbreviations.txt").

        Returns:
            Deep copy of the dictionary with parsed abbreviations.

        Raises:
            FileNotFound: If the specified file is not found.
        """
        records_dict_copy = (
            copy.deepcopy(records_dict) if records_dict is not None else {}
        )
        file_path = Path(path) / abbr_file
        if not file_path.exists():
            raise FileNotFound(f"{file_path} not found")

        with open(file_path, "r") as file:
            for line_index, line in enumerate(file, start=1):
                if Record.regex_abbr.match(line):
                    abbr, driver, team = line.strip().split("_")
                    records_dict_copy[abbr] = Record(driver=driver, team=team)
                else:
                    abbr_error = f"Bad data: {abbr_file}, line: {line_index}"
                    records_dict_copy[abbr_error] = Record(error=abbr_error)
        return records_dict_copy

    @staticmethod
    def read_logs(
        path: str | Path,
        records_dict: dict[str, "Record"] = None,
        start_file: str = "start.log",
        end_file: str = "end.log",
    ) -> dict[str, "Record"]:
        """
        Validate the file paths, extract start and end data, and save a deep copy of the data
        in dictionary if record keys are available, retaining lines that match the reg. ex. pattern.

        Args:
            path: Path to the directory containing logs files.
            records_dict: Optional dictionary to populate (defaults to empty).
            start_file: Name of the start file (defaults to "start.log").
            end_file: Name of the end file (defaults to "end.log").

        Returns:
            Deep copy of the dictionary with parsed start and end logs.

        Raises:
            FileNotFound: If the specified start or end file is not found.
        """
        records_dict_copy = (
            copy.deepcopy(records_dict) if records_dict is not None else {}
        )

        file_info = [
            (start_file, "start", "start", datetime.fromisoformat),
            (end_file, "end", "end", datetime.fromisoformat),
        ]

        for file_name, log_type, record_attr, parse_time_func in file_info:
            file_path = Path(path) / file_name
            if not file_path.exists():
                raise FileNotFound(f"{file_path} not found")

            with open(file_path, "r") as file:
                for line_index, line in enumerate(file, start=1):
                    if match := re.match(Record.regex_log, line.strip()):
                        record_key, log_time = match.groups()
                        if record_key in records_dict_copy:
                            record = records_dict_copy[record_key]
                            if not record.start and record_attr == "start":
                                record.start = parse_time_func(log_time)
                            elif not record.end and record_attr == "end":
                                record.end = parse_time_func(log_time)
                        else:
                            log_error = f"Bad abbr: {record_key}, line: {line_index}, file: {file_name}"
                            records_dict_copy[log_error] = Record(error=log_error)
                    else:
                        log_error = f"Bad data: {file_name}, line: {line_index}"
                        records_dict_copy[log_error] = Record(error=log_error)

        return records_dict_copy

    @staticmethod
    def build_report(
        path: str | Path,
        order: str = "asc",
    ) -> tuple[list["Record"], list["Record"]]:
        """
        Compiles a report summarizing valid and invalid records based on provided logs and abbreviations.

        Reads log files (end.log and start.log) and parses abbreviations from a text file (abbreviations.txt).
        Validates the order operator and sorts the data in ascending or descending order.

        Args:
            path: Path to the directory containing the log files and abbreviations text file.
            order: Sort order for records, 'asc' (ascending) or 'desc' (descending). Defaults to 'asc'.

        Returns:
            Tuple with two lists: good_record as valid list and bad_records as list of broken records.
        """
        good_records, bad_records = [], []

        try:
            if order not in ["asc", "desc"]:
                raise ReportInvalidOrder(
                    f"Invalid sorting order '{order}'. Valid options are: asc, desc"
                )

            records_dict = Record.read_abbr(path=path)
            records_dict = Record.read_logs(path=path, records_dict=records_dict)

            for rec in records_dict.values():
                (bad_records if rec.error else good_records).append(rec)

            multiplier = 1 if order == "asc" else -1
            good_records.sort(key=lambda x: multiplier * (x.result or timedelta(0)))

        except ReportInvalidOrder as e:
            print(str(e))
            return [], []

        return good_records, bad_records

    @staticmethod
    def print_report(good_records, bad_records, border_line: int = 15):
        """
        Print a formatted report based on a list of ordered results.

        Args:
            good_records - list of valid records in dict
            bad_records - list of invalid records in dict
        """

        result_string = ""
        line_count = 0

        for num, rec in enumerate(good_records + bad_records, start=1):
            driver = rec.driver if rec.driver is not None else "Missing Data"
            team = rec.team if rec.team is not None else "Missing Data"
            result = rec.result if not rec.error else rec.error

            record_str = f"{num:2}. {driver:<20} | {team:<30} | {result}"

            result_string += record_str + "\n"

            if num == border_line:
                result_string += "-" * 75 + "\n"
            line_count += 1

        return result_string

    @classmethod
    def cli(cls, args_list=None) -> argparse.Namespace:
        """
        Parses command-line arguments and orchestrates the report generation process.

        Args:
            args_list (list): List of strings representing command-line arguments.

        Returns:
            argparse.Namespace: Parsed command-line arguments.

        Example:
            Record.cli(["--files", "../data", "--asc", "--driver", "Sergey Sirotkin"])
        """
        parser = argparse.ArgumentParser(description="Build Monaco report")

        parser.add_argument("--files", required=True, help="Path to folder")

        group = parser.add_mutually_exclusive_group()
        group.add_argument("--asc", action="store_true", help="Ascending order sort")
        group.add_argument("--desc", action="store_true", help="Descending order sort")
        parser.add_argument("--driver", help="Show statistics for a specific driver")

        args = parser.parse_args(args_list)

        args.sort = "asc" if args.asc else "desc"

        return args


if __name__ == "__main__":
    args = Record.cli()

    good_records, bad_records = Record.build_report(args.files, args.sort)

    if args.driver:
        good_records = [rec for rec in good_records if rec.driver == args.driver]
        bad_records = [rec for rec in bad_records if rec.driver == args.driver]

    print(Record.print_report(good_records, bad_records))
